package com.afeiamic.springcloud.lb;

import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.client.ServiceInstance;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

@Component
@Slf4j
public class MyLB implements LoadBalancer {

    private final AtomicInteger nextIdx = new AtomicInteger(0);

    private int getAndIncrement() {
        int next, cur;
        do {
            cur = this.nextIdx.get();
            next = (cur == Integer.MAX_VALUE ? 0 : cur + 1);
        } while (!this.nextIdx.compareAndSet(cur, next));
        log.info("==========> 第 {} 次访问 <========= ", next);
        return next;
    }

    // 负载均衡算法：rest接口第几次请求数 % 服务器集群总数量 = 实际调用服务器位置下标，每次服务重启后rest接口计数从1开始
    @Override
    public ServiceInstance instances(List<ServiceInstance> serviceInstances) {
        int index = getAndIncrement() % serviceInstances.size();
        return serviceInstances.get(index);
    }
}
