package com.afeiamic.springcloud.controller;

import com.afeiamic.springcloud.entities.Payment;
import com.afeiamic.springcloud.entities.R;
import com.afeiamic.springcloud.service.PaymentService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cloud.client.ServiceInstance;
import org.springframework.cloud.client.discovery.DiscoveryClient;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import java.util.List;
import java.util.concurrent.TimeUnit;

@RestController
@Slf4j
public class PaymentController {

    @Resource
    private PaymentService paymentService;
    @Resource
    private DiscoveryClient discoveryClient;

    @Value("${server.port}")
    private String serverPort;

    @PostMapping(value = "/payment/create")
    public R<Integer> create(@RequestBody Payment payment) {
        int ret = paymentService.create(payment);
        log.info("========插入日志: " + ret);
        if (ret > 0) {
            return new R<>(200, "新增成功, serverPort: " + serverPort, ret);
        }
        return new R<>(555, "新增失败");
    }

    @GetMapping(value = "/payment/get/{id}")
    public R<Payment> getPaymentById(@PathVariable("id") Long id) {
        Payment payment = paymentService.getPaymentById(id);
        log.info("========查询结果: " + payment + "O(∩_∩)O哈哈~");
        if (payment != null) {
            return new R<>(200, "查询成功, serverPort: " + serverPort, payment);
        }
        return new R<>(444, "查询失败, 没有对应记录, id: 【" + id + "】");
    }

    @GetMapping(value = "/payment/discovery")
    public Object discovery() {
        // 获取并打印每一个服务名称
        List<String> services = discoveryClient.getServices();
        services.forEach(log::info);

        List<ServiceInstance> instances = discoveryClient.getInstances("CLOUD-PAYMENT-SERVICE");
        for (ServiceInstance instance : instances) {
            log.info(instance.getServiceId() + "\t" + instance.getHost() + "\t" + instance.getPort() + "\t" + instance.getUri());
        }
        return this.discoveryClient;
    }

    @GetMapping(value = "/payment/lb")
    public String getPaymentLB() {
        return serverPort;
    }

    @GetMapping(value = "/payment/feign/timeout")
    public String paymentFeignTimeout() {
        // 暂停几秒钟线程
        try { TimeUnit.SECONDS.sleep(3); } catch (InterruptedException e) { e.printStackTrace();}
        return serverPort;
    }

    @GetMapping("/payment/zipkin")
    public String paymentZipkin() {
        return "hi ,I am paymentZipkin server fall back，welcome to my home8001，O(∩_∩)O哈哈~";
    }
}
