package com.afeiamic.springcloud.controller;

import com.afeiamic.springcloud.entities.Payment;
import com.afeiamic.springcloud.entities.R;
import com.afeiamic.springcloud.service.PaymentService;
import com.alibaba.csp.sentinel.annotation.SentinelResource;
import com.alibaba.csp.sentinel.slots.block.BlockException;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.client.RestTemplate;

import javax.annotation.Resource;

@RestController
public class CircleBreakerController {

    private static final String SERVER_URL = "http://nacos-payment-provider";

    @Resource
    private RestTemplate restTemplate;

    @SuppressWarnings({"unchecked"})
    @GetMapping("/consumer/fallback/{id}")
    // @SentinelResource(value = "fallback", fallback = "handleFallback") // fallback 负责处理业务异常
    // @SentinelResource(value = "fallback", blockHandler = "blockHandler") // blockHandler 负责在 sentinel 里面配置的降级限流
    @SentinelResource(value = "fallback", blockHandler = "blockHandler", fallback = "handleFallback",
            exceptionsToIgnore = IllegalArgumentException.class) // 加入报该异常不再有 fallback 兜底, 没有降级效果
    public R<Payment> fallback(@PathVariable Long id) {
        R<Payment> result = restTemplate.getForObject(SERVER_URL + "/paymentSQL/" + id, R.class);
        if (id == 4) {
            throw new IllegalArgumentException("IllegalArgumentException, 非法参数异常！");
        } else {
            assert result != null;
            if (result.getData() == null) {
                throw new NullPointerException("NullPointerException, 该ID没有对应记录,空指针异常");
            }
        }
        return result;
    }
    public R<Payment> handleFallback(@PathVariable Long id, Throwable e) {
        Payment payment = new Payment(id, null);
        return new R<>(444, "兜底异常 handleFallback,exception 内容" + e.getMessage(), payment);
    }
    public R<Payment> blockHandler(@PathVariable Long id, BlockException e) {
        Payment payment = new Payment(id, null);
        return new R<>(445, "blockHandler-sentinel 限流 , 无此流水 : blockException" + e.getMessage(), payment);
    }

    //==================OpenFeign
    @Resource
    private PaymentService paymentService;

    @GetMapping("/consumer/paymentSQL/{id}")
    public R<Payment> paymentSQL(@PathVariable Long id) {
        if (id == 4) {
            throw new RuntimeException("没有该id");
        }
        return paymentService.paymentSQL(id);
    }
}
