package com.afeiamic.springcloud.controller;

import com.afeiamic.springcloud.entities.Payment;
import com.afeiamic.springcloud.entities.R;
import com.afeiamic.springcloud.myhandler.CustomerBlockHandler;
import com.afeiamic.springcloud.service.FlowLimitService;
import com.alibaba.csp.sentinel.annotation.SentinelResource;
import com.alibaba.csp.sentinel.slots.block.BlockException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import javax.annotation.Resource;
import java.util.concurrent.TimeUnit;

@RestController
@Slf4j
public class FlowLimitController {

    @Resource
    private FlowLimitService flowLimitService;

    @GetMapping(value = "/testA")
    public String testA() {
        return "------testA";
    }

    @GetMapping(value = "/testB")
    public String testB() {
        return "------testB";
    }

    @GetMapping(value = "/testC")
    public String testC() {
        return flowLimitService.name() + "from testC";
    }

    @GetMapping(value = "/testD")
    public String testD() {
        return flowLimitService.name() + "from testD";
    }

    @GetMapping(value = "/testE")
    public String testE() {
        // 暂停几秒钟线程
        try { TimeUnit.SECONDS.sleep(1); } catch (InterruptedException e) { e.printStackTrace();}
        log.info("=======> testE 测试 RT <=======");
        return "-------testE";
    }

    @GetMapping(value = "/testF")
    public String testF() {
        log.info("testF 测试异常比例数");
        int age = 10 / 0;
        return "------testF";
    }

    @GetMapping(value = "/testG")
    public String testG() {
        log.info("testG 测试异常数");
        int age = 10 / 0;
        return "------testG";
    }

    @GetMapping("/testHotKey")
    @SentinelResource(value = "hotKey", blockHandler = "hotKeyHandler")
    public String testHotKey(
            @RequestParam(value = "p1", required = false) String p1,
            @RequestParam(value = "p2", required = false) String p2) {
        return "hotKey";
    }
    public static String hotKeyHandler(@RequestParam(value = "p1", required = false) String p1,
                                       @RequestParam(value = "p2", required = false) String p2, BlockException e) {
        return "hostKey request faster";
    }

    @GetMapping(value = "byResource")
    @SentinelResource(value = "byResource", blockHandler = "handleException")
    public R<Payment> byResource() {
        return new R<>(200, "按资源名称限流测试 OK", new Payment(2023L, "serial001"));
    }
    public R<Payment> handleException(BlockException exception) {
        return new R<>(444, exception.getClass().getCanonicalName() + "\t服务不可用");
    }

    @GetMapping( value = "/rateLimit/byUrl")
    @SentinelResource(value = "byUrl")
    public R<Payment> retaLimit() {
        return new R<>(200, "/rateLimit/byUrl successful!", new Payment(2023L, "serial002"));
    }

    /**
     * 自定义通用的限流处理逻辑
     *
     * blockHandlerClass = CustomerBlockHandler.class
     * blockHandler = handleException2
     * 上述配置：找 CustomerBlockHandler 类里的 handleException2 方法进行兜底处理
     */
    @GetMapping("/rateLimit/customerBlockHandler")
    @SentinelResource(value = "customerBlockHandler", blockHandlerClass = CustomerBlockHandler.class, blockHandler = "handleException")
    public R<Payment> customerBlockHandler() {
        return new R<>(200, " 按客户自定义限流处理逻辑 ");
    }
}
