package com.afeiamic.springcloud.service.impl;

import com.afeiamic.springcloud.dao.OrderDao;
import com.afeiamic.springcloud.domain.Order;
import com.afeiamic.springcloud.service.AccountService;
import com.afeiamic.springcloud.service.OrderService;
import com.afeiamic.springcloud.service.StorageService;
import io.seata.spring.annotation.GlobalTransactional;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;

@Service
@Slf4j
public class OrderServiceImpl implements OrderService {

    @Resource
    private OrderDao orderDao;
    @Resource
    private StorageService storageService;
    @Resource
    private AccountService accountService;

    @GlobalTransactional(name = "fsp-create-order", rollbackFor = Exception.class)
    @Override
    public void create(Order order) {
        log.info("==================> 【订单微服务】新建订单(开始)......");

        // 创建订单
        orderDao.create(order);

        // 扣减库存
        log.info("==================> 【订单微服务】远程调用【库存微服务】, 扣减库存(开始)......");
        storageService.decrease(order.getProductId(), order.getCount());
        log.info("==================> 【订单微服务】远程调用【库存微服务】, 扣减库存(结束)......");

        // 扣减账户余额
        log.info("==================> 【订单微服务】远程调用【账户微服务】, 扣减账户余额(开始)......");
        accountService.decrease(order.getUserId(), order.getMoney());
        log.info("==================> 【订单微服务】远程调用【账户微服务】, 扣减账户余额(结束)......");

        // 修改订单状态
        log.info("==================> 【订单微服务】修改订单状态(开始)......");
        orderDao.update(order.getUserId(), 0);
        log.info("==================> 【订单微服务】修改订单状态(结束)......");

        log.info("==================> 【订单微服务】新建订单(结束)......");
    }
}
